// Copyright (c) 1999, 2000
//      Maurice Castro (maurice@serc.rmit.edu.au)
//      All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 
// This software is provided by the original author and contributors ``as
// is'' and any express or implied warranties, including, but not limited
// to, the implied warranties of merchantability and fitness for a
// particular purpose are disclaimed.  In no event shall the authors or
// contributors be liable for any direct, indirect, incidental, special,
// exemplary, or consequential damages (including, but not limited to,
// procurement of substitute goods or services; loss of use, data, or
// profits; or business interruption) however caused and on any theory of
// liability, whether in contract, strict liability, or tort (including
// negligence or otherwise) arising in any way out of the use of this
// software, even if advised of the possibility of such damage.
//
// $Id: ftp.java,v 1.5 2000/01/21 07:45:37 maurice Exp $


package ftp;

// partially impletents RFC 959
// has some unix foibles
// provides both passive and active connections

// Maurice Castro
// Fri Nov 12 14:24:15 EST 1999

import java.io.*;
import java.lang.*;
import java.util.*;
import java.net.*;
import lclhst;

public class ftp
{
// constants
	static final private int CMD_PORT = 21;
	static final private int DATA_PORT = 20;
	static final private char REP_POS_PRELIMINARY = '1';
	static final private char REP_POS_COMPLETION = '2';
	static final private char REP_POS_INTERMEDIARY = '3';
	static final private char REP_TRANSIENT_NEG_COMPLETION = '4';
	static final private char REP_PERMANENT_NEG_COMPLETION = '5';
	static final private int PASSIVE = 1;
	static final private int ACTIVE = 0;
	static final private int BINARY = 3;
	static final private int ASCII = 2;
	static final private int PASSIVE_TIME_OUT_MS = 10000;
	static final private int RETRY_DELAY_MS = 500;

// member data 
	private Socket m_cmdSock;
	private BufferedReader m_cmdReader;
	private PrintWriter m_cmdWriter;
	private String m_response;
	private String m_responseCode;
	private ServerSocket m_dataPort;
	private Socket m_dataSock;
	private BufferedReader m_bufRead;
	private BufferedWriter m_bufWrite;
	private InputStream m_inStr;
	private OutputStream m_outStr;
	private int connectMode = ACTIVE;
	private int xlateMode = ASCII;
	public int debug = 0;

// A psion / epoc workaround for a localhost bug
	lclhst lhost = new lclhst();

// open and close the connection
	public void open(String host) throws IOException, UnknownHostException
	{
		m_cmdSock = new Socket(host, CMD_PORT);
		m_cmdReader = new BufferedReader(new InputStreamReader(m_cmdSock.getInputStream()));
		m_cmdWriter = new PrintWriter(m_cmdSock.getOutputStream(), true);
		getResponse();
	}

	public void close() throws IOException
	{
		m_cmdSock.close();
	}

// ftp functions
	public void user(String user) throws IOException
	{
		sendCmd("USER " + user);
	}

	public void pass(String pass) throws IOException
	{
		sendCmd("PASS " + pass);
	}

	public void cd(String path) throws IOException
	{
		sendCmd("CWD " + path);
	}

	public void rmdir(String path) throws IOException
	{
		sendCmd("RMD " + path);
	}

	public void delete(String path) throws IOException
	{
		sendCmd("DELE " + path);
	}

	public void mkdir(String path) throws IOException
	{
		sendCmd("MKD " + path);
	}
	
	public void pwd() throws IOException
	{
		sendCmd("PWD");
	}

	public void cdup() throws IOException
	{
		sendCmd("CDUP");
	}

	public void status() throws IOException
	{
		sendCmd("STAT");
	}

	public void structMount(String path) throws IOException
	{
		sendCmd("SMNT " + path);
	}

	public void reinitialize() throws IOException
	{
		sendCmd("CDUP");
	}

	public void logout() throws IOException
	{
		sendCmd("QUIT");
	}

	public void noop() throws IOException
	{
		sendCmd("NOOP");
	}

	public void ascii() 
	{
		xlateMode = ASCII;
	}

	public void binary() 
	{
		xlateMode = BINARY;
	}

	public void abort() throws IOException
	{
		// as java does not provide OOB connection this
		// function is not likely to be very useful
		sendCmd("ABOR");
	}

	public BufferedReader list(String path) throws IOException
	{
		setAscii();
		if (path == null)
			sendCmd("NLST");
		else
			sendCmd("NLST " + path);
		if (!isPositivePreliminary())
			return null;
		dataConn();
		m_bufRead = new BufferedReader(new InputStreamReader(m_dataSock.getInputStream()));
		return(m_bufRead);
	}

	public BufferedReader list() throws IOException
	{
		return(list(null));
	}

	public BufferedReader dir(String path) throws IOException
	{
		setAscii();
		if (path == null)
			sendCmd("LIST");
		else
			sendCmd("LIST " + path);
		if (!isPositivePreliminary())
			return null;
		dataConn();
		m_bufRead = new BufferedReader(new InputStreamReader(m_dataSock.getInputStream()));
		return(m_bufRead);
	}

	public BufferedReader dir() throws IOException
	{
		return(dir(null));
	}

	public InputStream retrieve(String path) throws IOException
	{
		setXlate();
		if (!isPositiveCompletion())
			return null;
		sendCmd("RETR " + path);
		if (!isPositivePreliminary())
			return null;
		dataConn();
		m_inStr = m_dataSock.getInputStream();
		return(m_inStr);
	}

	public OutputStream store(String path) throws IOException
	{
		setXlate();
		if (!isPositiveCompletion())
			return null;
		sendCmd("STOR " + path);
		if (!isPositivePreliminary())
			return null;
		dataConn();
		m_outStr = m_dataSock.getOutputStream();
		return(m_outStr);
	}

// file translations
	private void setXlate() throws IOException
	{
		switch (xlateMode)
		{
			case BINARY: setBinary(); break;
			case ASCII: setAscii(); break;
		}
	}

	private void setAscii() throws IOException
	{
		sendCmd("TYPE A N");
	}

	private void setBinary() throws IOException
	{
		sendCmd("TYPE I");
	}

// response code functions
	public String getMessage()
	{
		return(m_response);
	}
	
	public String toString()
	{
		return(m_response);
	}

	public String getReturnCode()
	{
		return m_responseCode;
	}

	public void getResponse() throws IOException
	{
		String response = m_cmdReader.readLine();
		if ((debug & 0x2) != 0)
			System.out.println("<" + response);
		m_response = response;
		m_responseCode = m_response.substring(0,3);
		if (response.charAt(3) == '-')
		{
			char res[] = new char[3];
			do
			{
				response = m_cmdReader.readLine();
				if ((debug & 0x2) != 0)
					System.out.println("<" + response);
				m_response = m_response + "\n" + response;
				// according to rfc answers end when a 
				// line containing same result code issued
				// and unix implementation
				// answers end before another result code
				// is issued
			} while (!(new Character('a')).isDigit(response.charAt(0)));
		}
	}

	public boolean isPositivePreliminary()
	{
		return m_responseCode.charAt(0) == REP_POS_PRELIMINARY;
	}

	public boolean isPositiveCompletion()
	{
		return m_responseCode.charAt(0) == REP_POS_COMPLETION;
	}

	public boolean isPositiveIntermediary()
	{
		return m_responseCode.charAt(0) == REP_POS_INTERMEDIARY;
	}

	public boolean isTransientNegativeCompletion()
	{
		return m_responseCode.charAt(0) == REP_TRANSIENT_NEG_COMPLETION;
	}

	public boolean isPermanentNegativeCompletion()
	{
		return m_responseCode.charAt(0) == REP_PERMANENT_NEG_COMPLETION;
	}

// send an ftp command 
	private void sendCmd(String s) throws IOException
	{
		m_cmdWriter.print(s + "\r\n");
		m_cmdWriter.flush();
		if ((debug & 0x1) != 0)
			System.out.println(">" + s);
		getResponse();
	}

// set passive / active mode connection
	public void passive() 
	{
		connectMode = PASSIVE;
	}

	public void active() 
	{
		connectMode = ACTIVE;
	}

// data sockets
	public void getDataSock() throws IOException
	{
		if (m_dataSock != null)
		{
			m_dataSock.close();
			m_dataSock = null;
		}
		int port;
		String inetAddr;
		if (connectMode == PASSIVE)
		{
			// because we are using stream mode (see RFC 959 pg 19)
			// we may have to retry the PASV command until the 
			// socket used for the last transfer has closed
			Date mark = new Date();
			sendCmd("PASV");
			while (isTransientNegativeCompletion())
			{
				try
				{
					Thread.sleep(RETRY_DELAY_MS);
				}
				catch (InterruptedException e)
				{
					// nothing to do if this happens
				}
				if (((new Date()).getTime() - mark.getTime()) > 					PASSIVE_TIME_OUT_MS) 
					break;
				sendCmd("PASV");
			}
			if (isPositiveCompletion())
			{
				String csl;
				csl = m_response.substring(m_response.indexOf('(') + 1, m_response.indexOf(')'));
				StringTokenizer tok = new StringTokenizer(csl, ",");
				if (tok.countTokens() != 6)
					return;
				inetAddr = tok.nextToken() + "." + tok.nextToken() + "." + tok.nextToken() + "." + tok.nextToken();
				port = (new Integer(tok.nextToken())).intValue() * 256;
				port = (new Integer(tok.nextToken())).intValue() + port;
				m_dataSock = new Socket(inetAddr, port);
			}
		}
		else
		{
			if (m_dataPort != null)
			{
				m_dataPort.close();
				m_dataPort = null;
			}
			m_dataPort = new ServerSocket(0);
			// The following works for Unix ... but needs a
			// Work around for the psion
			// String host = InetAddress.getLocalHost().getHostAddress().replace('.', ',');
			String host = lhost.localhost().replace('.', ',');
			String portStr = m_dataPort.getLocalPort() / 256 + "," + m_dataPort.getLocalPort() % 256;
			sendCmd("PORT " + host + "," + portStr); 
		}
		return;
	}

	public void closeDataSock() throws IOException
	{
		if (m_bufRead != null)
		{
			m_bufRead.close();
			m_bufRead = null;
		}
		if (m_inStr != null)
		{
			m_inStr.close();
			m_inStr = null;
		}
		if (m_outStr != null)
		{
			m_outStr.close();
			m_outStr = null;
		}
		if (m_bufWrite != null)
		{
			m_bufWrite.close();
			m_bufWrite = null;
		}
		if (m_dataSock != null)
		{
			m_dataSock.close();
			m_dataSock = null;
		}
	}

	private void dataConn() throws IOException
	{
		if (connectMode == PASSIVE)
		{
			// do nothing
		}
		else
		{
			m_dataSock = m_dataPort.accept();
		}
	}
}
